#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from typing import Any, List
import numpy as np

from renormalizer.model import Op, OpSum, Model
from renormalizer import BasisHalfSpin, BasisSimpleElectron, BasisMultiElectron, BasisMultiElectronVac, Mps
from renormalizer.model.basis import BasisSet
import tensorcircuit as tc

from tencirchem.constants import DISCARD_EPS


# to be added to new DOFs
DOF_SALT = "TCCQUBIT"


def check_basis_type(basis):
    for b in basis:
        if isinstance(b, (BasisMultiElectronVac,)):
            raise TypeError(f"Unsupported basis: {type(b)}")
        if isinstance(b, BasisMultiElectron) and len(b.dofs) != 2:
            raise ValueError(f"For only two DOFs are allowed in BasisMultiElectron. Got {b}")


def qubit_encode_op(terms, basis, boson_encoding=None):
    check_basis_type(basis)
    if isinstance(terms, Op):
        terms = [terms]

    model = Model(basis, terms)

    old_ham_terms = []
    # `reversed` is for normal ordering with `pop`
    for op in reversed(terms):
        old_ham_terms.append(op.split_elementary(model.dof_to_siteidx))

    new_ham_terms = []
    while old_ham_terms:
        terms, factor = old_ham_terms.pop()
        opsum_list = []
        for op in terms:
            opsum = transform_op(op, model.dof_to_basis[op.dofs[0]], boson_encoding)
            opsum_list.append(opsum)

        new_term = 1
        for opsum in opsum_list:
            new_term = new_term * opsum
        new_term = new_term * factor

        new_ham_terms.extend(new_term)

    # post process
    # pick out constants
    identity_terms = []
    non_identity_terms = OpSum()
    for op in new_ham_terms:
        if op.is_identity:
            identity_terms.append(op)
        else:
            non_identity_terms.append(op.squeeze_identity())

    constant = sum([op.factor for op in identity_terms])

    return non_identity_terms.simplify(atol=DISCARD_EPS), constant


def qubit_encode_basis(basis: List[BasisSet], boson_encoding=None):
    spin_basis = []
    for b in basis:
        if isinstance(b, BasisMultiElectron):
            assert b.nbas == 2
            spin_basis.append(BasisHalfSpin(b.dofs))
        elif b.is_phonon:
            if boson_encoding is None:
                new_dofs = [b.dof]
            elif boson_encoding == "unary":
                new_dofs = [(b.dof, f"{DOF_SALT}-{i}") for i in range(b.nbas)]
            else:
                assert boson_encoding.lower() in ["binary", "gray"]
                n_qubits = int(np.ceil(np.log2(b.nbas)))
                new_dofs = [(b.dof, f"{DOF_SALT}-{i}") for i in range(n_qubits)]
            new_basis = [BasisHalfSpin(dof) for dof in new_dofs]
            spin_basis.extend(new_basis)
        else:
            spin_basis.append(BasisHalfSpin(b.dof))

    return spin_basis


def transform_op(op: Op, basis: BasisSet, boson_encoding: str = None) -> OpSum:
    # `op` is "elementary": all terms are in the `basis`
    assert op.factor == 1

    if set(op.split_symbol) == {"I"}:
        return OpSum([op])

    if isinstance(basis, (BasisHalfSpin, BasisSimpleElectron, BasisMultiElectron)):
        if isinstance(basis, BasisMultiElectron):
            assert len(basis.dof) == 2
            new_dof = basis.dofs
        else:
            new_dof = op.dofs[0]
        return transform_op_direct(op, new_dof, basis)

    # in principle can encode basis sets such as BasisMultiElectron,
    # but I think it's unnatural and not necessary
    assert basis.is_phonon
    return transform_op_boson(op, basis, boson_encoding)


def get_elem_qubit_op_direct(row_idx: int, col_idx: int, dof: Any):
    if (row_idx, col_idx) == (0, 0):
        return 1 / 2 * (Op("I", dof) + Op("Z", dof))
    elif (row_idx, col_idx) == (0, 1):
        return 1 / 2 * (Op("X", dof) + 1j * Op("Y", dof))
    elif (row_idx, col_idx) == (1, 0):
        return 1 / 2 * (Op("X", dof) - 1j * Op("Y", dof))
    else:
        assert (row_idx, col_idx) == (1, 1)
        return 1 / 2 * (Op("I", dof) - Op("Z", dof))


def transform_op_direct(op: Op, dof: Any, basis: BasisSet):
    if basis.nbas != 2:
        raise ValueError("Direct encoding only support two level basis")
    mat = basis.op_mat(op)
    ret = OpSum()
    for row_idx, col_idx in zip(*np.nonzero(mat)):
        ret += mat[row_idx, col_idx] * get_elem_qubit_op_direct(row_idx, col_idx, dof)
    return ret.simplify(atol=DISCARD_EPS)


def get_elem_qubit_op_unary(row_idx: int, col_idx: int, new_dofs: List[Any]):
    if row_idx == col_idx:
        dof_list = [new_dofs[row_idx]]
        return 1 / 2 * (Op("I", dof_list) - Op("Z", dof_list))
    else:
        des = 1 / 2 * (Op("X", new_dofs[col_idx]) + 1j * Op("Y", new_dofs[col_idx]))
        cre = 1 / 2 * (Op("X", new_dofs[row_idx]) - 1j * Op("Y", new_dofs[row_idx]))
        # exploit the commutation property for simplification
        if new_dofs[row_idx] < new_dofs[col_idx]:
            return cre * des
        else:
            return des * cre


def transform_op_boson_unary(op: Op, dof: Any, basis: BasisSet):
    new_dofs = [(dof, f"{DOF_SALT}-{i}") for i in range(basis.nbas - 1, -1, -1)]
    mat = basis.op_mat(op)
    ret = OpSum()
    for row_idx, col_idx in zip(*np.nonzero(mat)):
        ret += mat[row_idx, col_idx] * get_elem_qubit_op_unary(row_idx, col_idx, new_dofs)
    return ret.simplify(atol=DISCARD_EPS)


def get_elem_qubit_op_binary(row_idx: int, col_idx: int, new_dofs: List[Any], code_strs: List[str]):
    # gray_code_ints = [int(s, base=2) for s in gray_code_strs]
    n_qubits = len(new_dofs)
    if row_idx == col_idx:
        op_list = []
        for i in range(n_qubits):
            dof = new_dofs[i]
            if code_strs[row_idx][i] == "0":
                new_op = 1 / 2 * (Op("I", dof) + Op("Z", dof))
            else:
                new_op = 1 / 2 * (Op("I", dof) - Op("Z", dof))
            op_list.append(new_op)
        return OpSum.product(op_list)
    else:
        # |code1><code2|
        code1 = code_strs[row_idx]
        code2 = code_strs[col_idx]
        # diff_idx = [i for i in bin(code1 ^ code2)[2:] if i]
        op_list = []
        for i in range(n_qubits):
            dof = new_dofs[i]
            if code1[i] == code2[i]:
                if code1[i] == "0":
                    new_op = 1 / 2 * (Op("I", dof) + Op("Z", dof))
                else:
                    new_op = 1 / 2 * (Op("I", dof) - Op("Z", dof))
            else:
                if code1[i] + code2[i] == "01":
                    new_op = 1 / 2 * (Op("X", dof) + 1j * Op("Y", dof))
                else:
                    new_op = 1 / 2 * (Op("X", dof) - 1j * Op("Y", dof))
            op_list.append(new_op)
        return OpSum.product(op_list)


def transform_op_boson_binary(op: Op, dof: Any, basis: BasisSet, encoding: str):
    n_qubits = (basis.nbas - 1).bit_length()
    new_dofs = [(dof, f"{DOF_SALT}-{i}") for i in range(n_qubits)]
    if encoding == "gray":
        code_strs = get_gray_codes(n_qubits)
    else:
        assert encoding == "binary"
        code_strs = get_binary_codes(n_qubits)

    mat = basis.op_mat(op)
    ret = OpSum()
    for row_idx, col_idx in zip(*np.nonzero(mat)):
        ret += mat[row_idx, col_idx] * get_elem_qubit_op_binary(row_idx, col_idx, new_dofs, code_strs)
    return ret.simplify(atol=DISCARD_EPS)


def transform_op_boson(op, basis, encoding=None):
    assert op.factor == 1

    if encoding is None:
        return transform_op_direct(op, op.dofs[0], basis)
    elif encoding == "unary":
        return transform_op_boson_unary(op, op.dofs[0], basis)
    elif encoding.lower() in ["binary", "gray"]:
        return transform_op_boson_binary(op, op.dofs[0], basis, encoding.lower())
    else:
        raise ValueError(f"Encoding '{encoding}' not supported")


def get_gray_codes(n):
    """Return n-bit Gray code in a list."""
    if n == 0:
        return [""]
    sub_gray_codes = get_gray_codes(n - 1)

    gray_codes0 = ["0" + code for code in sub_gray_codes]
    gray_codes1 = ["1" + code for code in reversed(sub_gray_codes)]

    return gray_codes0 + gray_codes1


def get_binary_codes(n):
    codes = [bin(i)[2:] for i in range(1 << n)]
    return ["0" * (n - len(code)) + code for code in codes]


def get_encoding(m, boson_encoding):
    if boson_encoding is None:
        assert m == 2
        encoding_order = "01"
    elif boson_encoding == "unary":
        encoding_order = ["0" * (m - 1 - i) + "1" + "0" * i for i in range(m)]
    elif boson_encoding == "binary":
        encoding_order = get_binary_codes((m - 1).bit_length())[:m]
    else:
        assert boson_encoding == "gray"
        encoding_order = get_gray_codes((m - 1).bit_length())[:m]
    return encoding_order


def get_init_circuit(model_ref, model, boson_encoding, init_condition):
    for k, v in init_condition.items():
        basis = model_ref.dof_to_basis[k]
        if not isinstance(v, int):
            if (
                isinstance(basis, BasisHalfSpin)
                and v.shape == (2, 2)
                and np.allclose(np.eye(2), v @ v.T.conj)
                and np.allclose(np.eye(2), v.T.conj @ v)
            ):
                continue
            else:
                return get_init_circuit_general(model_ref, model, boson_encoding, init_condition)

    # replace the dof_name key to site_index key
    circuit = tc.Circuit(len(model.basis))
    for k, v in init_condition.items():
        basis = model_ref.dof_to_basis[k]
        if isinstance(basis, BasisHalfSpin):
            if v == 1:
                circuit.X(model.dof_to_siteidx[k])
            elif v.shape == (2, 2):
                circuit.ANY(idx, unitary=v)
            else:
                assert v == 0
        elif isinstance(basis, BasisMultiElectron):
            if v == 1:
                idx = model.dof_to_siteidx[basis.dofs]
                circuit.X(idx)
            else:
                assert v == 0
        else:
            assert basis.is_phonon
            if boson_encoding is None:
                assert v in [0, 1]
                target = [v]
            elif boson_encoding == "unary":
                target = [0] * len(basis.nbas)
                target[len(basis.nbas) - v - 1] = 1
            elif boson_encoding == "binary":
                target = get_binary_codes((basis.nbas - 1).bit_length())[v]
            else:
                assert boson_encoding == "gray"
                target = get_gray_codes((basis.nbas - 1).bit_length())[v]
            for i, t in enumerate(target):
                if t == 1:
                    circuit.X(model.dof_to_siteidx[(k, f"{DOF_SALT}-{i}")])
    return circuit


def get_init_circuit_general(model_ref, model, boson_encoding, init_condition):
    mps = Mps.hartree_product_state(model_ref, init_condition)
    mps_state = mps.todense()
    subspace_idx = get_subspace_idx(model_ref.basis, boson_encoding)
    assert len(mps_state) == len(subspace_idx)
    n_qubits = len(model.basis)
    state = np.zeros(1 << n_qubits)
    state[subspace_idx] = mps_state
    return tc.Circuit(n_qubits, inputs=state)


def get_subspace_idx(basis_list, boson_encoding):
    subspace_idx = [""]
    for basis in basis_list:
        if isinstance(basis, (BasisSimpleElectron, BasisMultiElectron, BasisHalfSpin)):
            new_idx = "01"
        else:
            new_idx = get_encoding(basis.nbas, boson_encoding)
        new_subspace_idx = []
        for idx1 in subspace_idx:
            for idx2 in new_idx:
                new_subspace_idx.append(idx1 + idx2)
        subspace_idx = new_subspace_idx
    return [int(i, base=2) for i in subspace_idx]



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from typing import Tuple, List

import numpy as np
from pyscf.gto.mole import Mole
from pyscf.cc.addons import spatial2spin
from pyscf.fci import cistring
from tensorcircuit import Circuit

from tencirchem.utils.misc import rdm_mo2ao
from tencirchem.static.ucc import UCC
from tencirchem.static.ci_utils import get_ci_strings
from tencirchem.static.evolve_tensornetwork import get_circuit_givens_swap


class PUCCD(UCC):
    """
    Run paired UCC calculation.
    The interfaces are similar to :class:`UCCSD <tencirchem.UCCSD>`.
    """

    # todo: more documentation here and make the references right.
    # separate docstring examples for a variety of methods, such as energy()
    # also need to add a few comment on make_rdm1/2
    # https://arxiv.org/pdf/2002.00035.pdf
    # https://arxiv.org/pdf/1503.04878.pdf

    def __init__(
        self,
        mol: Mole,
        init_method: str = "mp2",
        active_space: Tuple[int, int] = None,
        mo_coeff: np.ndarray = None,
        engine: str = None,
        run_hf: bool = True,
        run_mp2: bool = True,
        run_ccsd: bool = True,
        run_fci: bool = True,
    ):
        r"""
        Initialize the class with molecular input.

        Parameters
        ----------
        mol: Mole
            The molecule as PySCF ``Mole`` object.
        init_method: str, optional
            How to determine the initial amplitude guess. Accepts ``"mp2"`` (default), ``"ccsd"``
            and ``"zeros"``.
        active_space: Tuple[int, int], optional
            Active space approximation. The first integer is the number of electrons and the second integer is
            the number or spatial-orbitals. Defaults to None.
        mo_coeff: np.ndarray, optional
            Molecule coefficients. If provided then RHF is skipped.
            Can be used in combination with the ``init_state`` attribute.
            Defaults to None which means RHF orbitals are used.
        engine: str, optional
            The engine to run the calculation. See :ref:`advanced:Engines` for details.
        run_hf: bool, optional
            Whether run HF for molecule orbitals. Defaults to ``True``.
        run_mp2: bool, optional
            Whether run MP2 for initial guess and energy reference. Defaults to ``True``.
        run_ccsd: bool, optional
            Whether run CCSD for initial guess and energy reference. Defaults to ``True``.
        run_fci: bool, optional
            Whether run FCI  for energy reference. Defaults to ``True``.

        See Also
        --------
        tencirchem.UCCSD
        tencirchem.KUPCCGSD
        tencirchem.UCC
        """
        super().__init__(
            mol,
            init_method,
            active_space,
            mo_coeff,
            hcb=True,
            engine=engine,
            run_hf=run_hf,
            run_mp2=run_mp2,
            run_ccsd=run_ccsd,
            run_fci=run_fci,
        )
        self.ex_ops, self.param_ids, self.init_guess = self.get_ex_ops(self.t1, self.t2)

    def get_ex1_ops(self, t1: np.ndarray = None):
        """Not applicable. Use :func:`get_ex_ops`."""
        raise NotImplementedError

    def get_ex2_ops(self, t2: np.ndarray = None):
        """Not applicable. Use :func:`get_ex_ops`."""
        raise NotImplementedError

    def get_ex_ops(self, t1: np.ndarray = None, t2: np.ndarray = None) -> Tuple[List[Tuple], List[int], List[float]]:
        """
        Get paired excitation operators for pUCCD ansatz.

        Parameters
        ----------
        t1: np.ndarray, optional
            Not used. Kept for consistency with the parent method.
        t2: np.ndarray, optional
            Not used. Kept for consistency with the parent method.

        Returns
        -------
        ex_op: List[Tuple]
            The excitation operators. Each operator is represented by a tuple of ints.
        param_ids: List[int]
            The mapping from excitations to parameters.
        init_guess: List[float]
            The initial guess for the parameters.

        Examples
        --------
        >>> from tencirchem import PUCCD
        >>> from tencirchem.molecule import h2
        >>> puccd = PUCCD(h2)
        >>> ex_op, param_ids, init_guess = puccd.get_ex_ops()
        >>> ex_op
        [(1, 0)]
        >>> param_ids
        [0]
        >>> init_guess  # doctest:+ELLIPSIS
        [...]
        """
        no, nv = self.no, self.nv
        if t2 is None:
            t2 = np.zeros((no, no, nv, nv))
        t2 = spatial2spin(t2)

        ex_ops = []
        ex_init_guess = []
        # to be consistent with givens rotation circuit
        for i in range(no):
            for a in range(nv - 1, -1, -1):
                ex_ops.append((no + a, i))
                ex_init_guess.append(t2[2 * i, 2 * i + 1, 2 * a, 2 * a + 1])
        return ex_ops, list(range(len(ex_ops))), ex_init_guess

    def make_rdm1(self, statevector=None, basis="AO"):
        __doc__ = super().make_rdm1.__doc__

        civector = self._statevector_to_civector(statevector)
        ci_strings = get_ci_strings(self.n_qubits, self.n_elec, self.hcb)

        n_active = self.n_qubits
        rdm1_cas = np.zeros([n_active] * 2)
        for i in range(n_active):
            bitmask = 1 << i
            arraymask = (ci_strings & bitmask) == bitmask
            value = float(civector @ (arraymask * civector))
            rdm1_cas[i, i] = 2 * value
        rdm1 = self.embed_rdm_cas(rdm1_cas)
        if basis == "MO":
            return rdm1
        else:
            return rdm_mo2ao(rdm1, self.hf.mo_coeff)

    def make_rdm2(self, statevector=None, basis="AO"):
        __doc__ = super().make_rdm2.__doc__

        civector = self._statevector_to_civector(statevector)
        ci_strings = get_ci_strings(self.n_qubits, self.n_elec, self.hcb)

        n_active = self.n_qubits
        rdm2_cas = np.zeros([n_active] * 4)
        for p in range(n_active):
            for q in range(p + 1):
                maskq = 1 << q
                maskp = 1 << p
                maskpq = maskp + maskq
                arraymask = (ci_strings & maskq) == maskq
                if p == q:
                    value = float(civector @ (arraymask * civector))
                else:
                    arraymask &= ((~ci_strings) & maskp) == maskp
                    excitation = ci_strings ^ maskpq
                    addr = cistring.strs2addr(n_active, self.n_elec // 2, excitation)
                    value = float(civector @ (arraymask * civector[addr]))

                rdm2_cas[p, q, p, q] = rdm2_cas[q, p, q, p] = value
                if p == q:
                    continue
                arraymask = (ci_strings & maskpq) == maskpq
                value = float(civector @ (arraymask * civector))

                rdm2_cas[p, p, q, q] = rdm2_cas[q, q, p, p] = 2 * value
                rdm2_cas[p, q, q, p] = rdm2_cas[q, p, p, q] = -value
        rdm2_cas *= 2
        rdm2 = self.embed_rdm_cas(rdm2_cas)
        # no need to transpose
        if basis == "MO":
            return rdm2
        else:
            return rdm_mo2ao(rdm2, self.hf.mo_coeff)

    def get_circuit(self, params=None, trotter=False, givens_swap=False) -> Circuit:
        """
        Get the circuit as TensorCircuit ``Circuit`` object.

        Parameters
        ----------
        params: Tensor, optional
            The circuit parameters. Defaults to None, which uses the optimized parameter.
            If :func:`kernel` is not called before, the initial guess is used.
        trotter: bool, optional
            Whether Trotterize the UCC factor into Pauli strings.
            Defaults to False.
        givens_swap: bool, optional
            Whether return the circuit with Givens-Swap gates.

        Returns
        -------
        circuit: :class:`tc.Circuit`
            The quantum circuit.
        """
        if not givens_swap:
            return super().get_circuit(params, trotter=trotter)
        else:
            params = self._check_params_argument(params, strict=False)
            return get_circuit_givens_swap(params, self.n_qubits, self.n_elec, self.init_state)

    @property
    def e_puccd(self):
        """
        Returns pUCCD energy
        """
        return self.energy()



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


import logging
from typing import Any, List, Tuple
from functools import partial

import numpy as np
from openfermion import jordan_wigner
import tensorcircuit as tc

from tencirchem.static.ci_utils import get_ci_strings, civector_to_statevector
from tencirchem.utils.backend import jit
from tencirchem.utils.misc import ex_op_to_fop, reverse_qop_idx
from tencirchem.utils.circuit import evolve_pauli, multicontrol_ry

logger = logging.getLogger(__name__)


Tensor = Any


def evolve_excitation(circuit: tc.Circuit, f_idx: tuple, qop, theta, hcb: bool, decompose_multicontrol: bool):
    n_qubits = circuit.circuit_param["nqubits"]
    f_idx = [n_qubits - 1 - idx for idx in f_idx]

    z_indices = []
    if not hcb:
        for idx, term in next(iter(qop.terms.keys())):
            if term != "Z":
                assert idx in f_idx
                continue
            z_indices.append(idx)

    def parity_gates(target, reverse=False):
        if len(z_indices) == 0:
            return
        if not reverse:
            for ii in range(len(z_indices) - 1):
                circuit.CNOT(z_indices[ii], z_indices[ii + 1])
            circuit.cz(z_indices[-1], target)
        else:
            circuit.cz(z_indices[-1], target)
            for ii in reversed(range(len(z_indices) - 1)):
                circuit.CNOT(z_indices[ii], z_indices[ii + 1])

    if len(f_idx) == 2:
        k, l = f_idx
        circuit.CNOT(k, l)
        parity_gates(k)
        circuit.cry(l, k, theta=theta)
        parity_gates(k, reverse=True)
        circuit.CNOT(k, l)
    else:
        assert len(f_idx) == 4
        k, l, i, j = f_idx
        circuit.CNOT(l, k)
        circuit.CNOT(j, i)
        circuit.CNOT(l, j)
        parity_gates(l)
        if not decompose_multicontrol:
            try:
                name = f"Ry({theta:.4f})"
            except TypeError:
                # jax tracer can't be formatted
                name = "Ry"
            circuit.multicontrol(i, j, k, l, ctrl=[0, 1, 0], unitary=tc.gates.ry_gate(theta).tensor, name=name)
        else:
            circuit.append(multicontrol_ry(theta), indices=[i, j, k, l])
        parity_gates(l, reverse=True)
        circuit.CNOT(l, j)
        circuit.CNOT(j, i)
        circuit.CNOT(l, k)

    return circuit


def get_init_circuit(n_qubits, n_elec, hcb, init_state=None, givens_swap=False):
    if init_state is None:
        # prepare HF state
        circuit = tc.Circuit(n_qubits)
        if not hcb:
            for i in range(n_elec // 2):
                circuit.X(n_qubits - 1 - i)
                circuit.X(n_qubits // 2 - 1 - i)
        else:
            if not givens_swap:
                for i in range(n_elec // 2):
                    circuit.X(n_qubits - 1 - i)
            else:
                for i in range(n_elec // 2):
                    circuit.X(i)
    elif isinstance(init_state, tc.Circuit):
        return init_state
    else:
        ci_strings = get_ci_strings(n_qubits, n_elec, hcb)
        statevector = civector_to_statevector(init_state, n_qubits, ci_strings)
        if givens_swap:
            statevector = statevector.reshape([2] * n_qubits)
            new_idx = list(range(n_qubits - n_elec // 2, n_qubits)) + list(range(n_qubits - n_elec // 2))
            statevector = statevector.transpose(new_idx).ravel()
        circuit = tc.Circuit(n_qubits, inputs=statevector)
    return circuit


def get_circuit(
    params,
    n_qubits,
    n_elec,
    ex_ops,
    param_ids,
    hcb: bool = False,
    init_state: tc.Circuit = None,
    decompose_multicontrol: bool = False,
    trotter: bool = False,
):
    if param_ids is None:
        assert len(params) == len(ex_ops)
        param_ids = list(range(len(params)))

    circuit = get_init_circuit(n_qubits, n_elec, hcb, init_state)

    for param_id, f_idx in zip(param_ids, ex_ops):
        theta = params[param_id]
        fop = ex_op_to_fop(f_idx, with_conjugation=True)
        qop = reverse_qop_idx(jordan_wigner(fop), n_qubits)
        if trotter:
            for pauli_string, v in qop.terms.items():
                if hcb:
                    pauli_string = [(idx, symbol) for idx, symbol in pauli_string if symbol != "Z"]
                circuit = evolve_pauli(circuit, pauli_string, -2 * v.imag * theta)
        else:
            # https://arxiv.org/pdf/2005.14475.pdf
            circuit = evolve_excitation(circuit, f_idx, qop, 2 * theta, hcb, decompose_multicontrol)

    return circuit


def get_gs_unitary(theta):
    # SWAP @ Givens Rotation
    a = [
        [1, 0, 0, 0],
        [0, -tc.backend.sin(theta), tc.backend.cos(theta), 0],
        [0, tc.backend.cos(theta), tc.backend.sin(theta), 0],
        [0, 0, 0, 1],
    ]
    return tc.backend.convert_to_tensor(a)


def get_gs_indices(no: int, nv: int) -> List[Tuple[int, int]]:
    """Givens-Swap indices"""
    layer1 = [np.array([no - 1, no])]
    for _ in range(nv - 1):
        layer1.append(layer1[-1] + 1)
    layer1 = np.array(layer1)
    ret = [layer1]
    for _ in range(no - 1):
        ret.append(ret[-1] - 1)
    ret = np.array(ret).reshape(-1, 2)
    assert len(ret) == no * nv
    return ret.tolist()


def get_circuit_givens_swap(params, n_qubits, n_elec, init_state=None):
    # https://arxiv.org/pdf/2002.00035.pdf
    # the swapped qubit index represents molecule orbitals 0 to n_qubits - 1
    # so we need to take negative of theta
    #             ┌───┐        ┌──────┐
    # q_0(MO 1) : ┤ X ├────────┤      ├─────────  MO 3
    #             ├───┤┌──────┐│  GS  │┌──────┐
    # q_1(MO 0) : ┤ X ├┤      ├┤ 3, 1 ├┤      ├─  MO 2
    #             └───┘│  GS  │├──────┤│  GS  │
    # q_2(MO 3) : ─────┤ 3, 0 ├┤      ├┤ 2, 1 ├─  MO 1
    #                  └──────┘│  GS  │└──────┘
    # q_3(MO 2) : ─────────────┤ 2, 0 ├─────────  MO 0
    #                          └──────┘
    circuit = get_init_circuit(n_qubits, n_elec, hcb=True, init_state=init_state, givens_swap=True)
    gs_indices = get_gs_indices(n_elec // 2, n_qubits - n_elec // 2)
    for i, (j, k) in enumerate(gs_indices):
        theta = params[i]
        unitary = get_gs_unitary(theta)
        try:
            name = f"Givens-SWAP({theta:.4f})"
        except TypeError:
            # jax tracer can't be formatted
            name = "Givens-SWAP"
        circuit.any(j, k, unitary=unitary, name=name)
    return circuit


@partial(jit, static_argnums=[1, 2, 3, 4, 5])
def get_statevector_tensornetwork(params, n_qubits, n_elec, ex_ops, param_ids, hcb=False, init_state=None):
    if tc.backend.name == "jax":
        logger.info(f"Entering `get_statevector_tensornetwork`. n_qubit: {n_qubits}")

    circuit = get_circuit(params, n_qubits, n_elec, ex_ops, param_ids, hcb, init_state)
    return circuit.state().real.reshape(-1)



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from functools import partial
import logging
from typing import Tuple

import numpy as np
from openfermion import jordan_wigner
import tensorcircuit as tc

from tencirchem.utils.backend import jit, fori_loop, scan, get_xp, get_uint_type
from tencirchem.utils.misc import ex_op_to_fop
from tencirchem.static.hamiltonian import apply_op
from tencirchem.static.ci_utils import get_ci_strings, get_addr, get_init_civector


logger = logging.getLogger(__name__)


def get_fket_permutation(f_idx, n_qubits, n_elec, ci_strings, strs2addr, hcb):
    mask = 0
    for i in f_idx:
        mask += 1 << i
    excitation = ci_strings ^ mask
    return get_addr(excitation, n_qubits, n_elec, strs2addr, hcb)


def get_fket_phase(f_idx, ci_strings):
    if len(f_idx) == 2:
        mask1 = 1 << f_idx[0]
        mask2 = 1 << f_idx[1]
    else:
        assert len(f_idx) == 4
        mask1 = (1 << f_idx[0]) + (1 << f_idx[1])
        mask2 = (1 << f_idx[2]) + (1 << f_idx[3])
    flip = ci_strings ^ mask1
    mask = mask1 | mask2
    masked = flip & mask
    positive = masked == mask
    negative = masked == 0
    return positive, negative


FERMION_PHASE_MASK_CACHE = {}


def get_fermion_phase(f_idx, n_qubits, ci_strings):
    if f_idx in FERMION_PHASE_MASK_CACHE:
        mask, sign = FERMION_PHASE_MASK_CACHE[f_idx]
    else:
        # fermion operator index, not sorted
        fop = ex_op_to_fop(f_idx)

        # pauli string index, already sorted
        qop = jordan_wigner(fop)
        mask_str = ["0"] * n_qubits
        for idx, term in next(iter(qop.terms.keys())):
            if term != "Z":
                assert idx in f_idx
                continue
            mask_str[n_qubits - 1 - idx] = "1"
        mask = get_uint_type()(int("".join(mask_str), base=2))

        if sorted(qop.terms.items())[0][1].real > 0:
            sign = -1
        else:
            sign = 1

        FERMION_PHASE_MASK_CACHE[f_idx] = mask, sign

    parity = ci_strings & mask
    assert parity.dtype in [np.uint32, np.uint64]
    if parity.dtype == np.uint32:
        mask = 0x11111111
        shift = 28
    else:
        mask = 0x1111111111111111
        shift = 60
    parity ^= parity >> 1
    parity ^= parity >> 2
    parity = (parity & mask) * mask
    parity = (parity >> shift) & 1

    return sign * np.sign(parity - 0.5).astype(np.int8)


def get_operators(n_qubits, n_elec, strs2addr, f_idx, ci_strings, hcb):
    if len(set(f_idx)) != len(f_idx):
        raise ValueError(f"Excitation {f_idx} not supported")
    xp = get_xp(tc.backend)
    fket_permutation = get_fket_permutation(f_idx, n_qubits, n_elec, ci_strings, strs2addr, hcb)
    fket_phase = xp.zeros(len(ci_strings))
    positive, negative = get_fket_phase(f_idx, ci_strings)
    fket_phase -= positive
    fket_phase += negative
    if not hcb:
        fket_phase *= get_fermion_phase(f_idx, n_qubits, ci_strings)
    f2ket_phase = xp.zeros(len(ci_strings))
    f2ket_phase -= positive
    f2ket_phase -= negative

    return fket_permutation, fket_phase, f2ket_phase


CI_OPERATOR_BATCH_CACHE = {}
CI_OPERATOR_CACHE = {}


@partial(jit, static_argnums=[0, 1, 2, 3])
def get_operator_tensors(n_qubits, n_elec, ex_ops, hcb=False):
    xp = get_xp(tc.backend)
    batch_key = (xp, tc.rdtypestr, n_qubits, n_elec, ex_ops, hcb)
    is_jax_backend = tc.backend.name == "jax"
    if not is_jax_backend and batch_key in CI_OPERATOR_BATCH_CACHE:
        return CI_OPERATOR_BATCH_CACHE[batch_key]

    ci_strings, strs2addr = get_ci_strings(n_qubits, n_elec, hcb, strs2addr=True)

    xp = get_xp(tc.backend)
    fket_permutation_tensor = xp.zeros((len(ex_ops), len(ci_strings)), dtype=get_uint_type())
    fket_phase_tensor = xp.zeros((len(ex_ops), len(ci_strings)), dtype=np.int8)
    f2ket_phase_tensor = xp.zeros((len(ex_ops), len(ci_strings)), dtype=np.int8)
    for i, f_idx in enumerate(ex_ops):
        if 64 < len(ex_ops):
            logger.info((i, f_idx))
        op_key = (xp, tc.rdtypestr, n_qubits, n_elec, f_idx, hcb)
        if not is_jax_backend and op_key in CI_OPERATOR_CACHE:
            fket_permutation, fket_phase, f2ket_phase = CI_OPERATOR_CACHE[op_key]
        else:
            fket_permutation, fket_phase, f2ket_phase = get_operators(
                n_qubits, n_elec, strs2addr, f_idx, ci_strings, hcb
            )
            if not is_jax_backend:
                CI_OPERATOR_CACHE[op_key] = fket_permutation, fket_phase, f2ket_phase
        fket_permutation_tensor[i] = fket_permutation
        fket_phase_tensor[i] = fket_phase
        f2ket_phase_tensor[i] = f2ket_phase

    fket_permutation_tensor = tc.backend.convert_to_tensor(fket_permutation_tensor)
    fket_phase_tensor = tc.backend.convert_to_tensor(fket_phase_tensor)
    f2ket_phase_tensor = tc.backend.convert_to_tensor(f2ket_phase_tensor)

    ret = ci_strings, fket_permutation_tensor, fket_phase_tensor, f2ket_phase_tensor
    if not is_jax_backend:
        CI_OPERATOR_BATCH_CACHE[batch_key] = ret
    return ret


@partial(jit, static_argnums=[1])
def get_theta_tensors(params, param_ids):
    theta_list = []
    for param_id in param_ids:
        theta_list.append(params[param_id])

    theta_tensor = tc.backend.convert_to_tensor(theta_list)
    theta_sin_tensor = tc.backend.sin(theta_tensor)
    theta_1mcos_tensor = 1 - tc.backend.cos(theta_tensor)
    return theta_sin_tensor, theta_1mcos_tensor


@jit
def evolve_civector_by_tensor(
    civector, fket_permutation_tensor, fket_phase_tensor, f2ket_phase_tensor, theta_sin, theta_1mcos
):
    def _evolve_excitation(j, _civector):
        _fket_phase = fket_phase_tensor[j]
        _fket_permutation = fket_permutation_tensor[j]
        fket = _civector[_fket_permutation] * _fket_phase
        f2ket = f2ket_phase_tensor[j] * _civector
        _civector += theta_1mcos[j] * f2ket + theta_sin[j] * fket
        return _civector

    return fori_loop(0, len(fket_permutation_tensor), _evolve_excitation, civector)


@partial(jit, static_argnums=[1, 2, 3, 4, 5])
def get_civector(params, n_qubits, n_elec, ex_ops, param_ids, hcb=False, init_state=None):
    if tc.backend.name == "jax":
        logger.info(f"Entering `get_civector`. n_qubit: {n_qubits}")

    ci_strings, fket_permutation_tensor, fket_phase_tensor, f2ket_phase_tensor = get_operator_tensors(
        n_qubits, n_elec, ex_ops, hcb
    )
    theta_sin, theta_1mcos = get_theta_tensors(params, param_ids)

    if init_state is None:
        civector = get_init_civector(len(ci_strings))
    else:
        civector = tc.backend.convert_to_tensor(init_state)
    civector = evolve_civector_by_tensor(
        civector, fket_permutation_tensor, fket_phase_tensor, f2ket_phase_tensor, theta_sin, theta_1mcos
    )

    return civector.reshape(-1)


def get_energy_and_grad_civector(
    params, hamiltonian, n_qubits, n_elec, ex_ops: Tuple, param_ids: Tuple, hcb: bool = False, init_state=None
):
    ket = get_civector(params, n_qubits, n_elec, ex_ops, param_ids, hcb, init_state)
    bra = apply_op(hamiltonian, ket)
    energy = bra @ ket
    # already cached
    op_tensors = get_operator_tensors(n_qubits, n_elec, ex_ops, hcb=hcb)
    theta_tensors = get_theta_tensors(params, param_ids)
    op_tensors = list(op_tensors) + list(theta_tensors)
    gradients_beforesum = _get_gradients_civector(bra, ket, *op_tensors[1:])

    gradients_beforesum = tc.backend.numpy(gradients_beforesum)
    gradients = np.zeros(params.shape)
    for grad, param_id in zip(gradients_beforesum, param_ids):
        gradients[param_id] += grad

    return energy, 2 * gradients


@jit
def _get_gradients_civector(
    bra, ket, fket_permutation_tensor, fket_phase_tensor, f2ket_phase_tensor, theta_sin, theta_1mcos
):
    scan_xs = fket_permutation_tensor, fket_phase_tensor, f2ket_phase_tensor, -theta_sin, theta_1mcos

    def _evolve_excitation(_civector, _fket_permutation, _fket_phase, _f2ket_phase):
        _civector += _f2ket_phase * _civector + _civector[_fket_permutation] * _fket_phase
        return _civector

    def get_grad(braket, scan_x):
        _bra, _ket = braket
        _fket_permutation, _fket_phase, _f2ket_phase, _theta_msin, _theta_1mcos = scan_x
        _ket = _evolve_excitation(_ket, _fket_permutation, _fket_phase * _theta_msin, _f2ket_phase * _theta_1mcos)
        _bra = _evolve_excitation(_bra, _fket_permutation, _fket_phase * _theta_msin, _f2ket_phase * _theta_1mcos)
        _fket = _ket[_fket_permutation] * _fket_phase
        grad = _bra @ _fket

        return (_bra, _ket), grad

    _, gradients = scan(get_grad, (bra, ket), scan_xs, len(fket_permutation_tensor), reverse=True)

    return gradients


def evolve_excitation_nocache(civector, fket_permutation, fket_phase, f2ket_phase, theta_1mcos, theta_sin):
    fket = civector[fket_permutation] * fket_phase
    f2ket = civector * f2ket_phase
    civector += theta_1mcos * f2ket + theta_sin * fket
    return civector


@partial(jit, static_argnums=[1, 2, 3, 4, 5])
def get_civector_nocache(params, n_qubits, n_elec, ex_ops, param_ids, hcb=False, init_state=None):
    if tc.backend.name == "jax":
        logger.info(f"Entering `get_civector_nocache`. n_qubit: {n_qubits}")

    theta_sin_tensor, theta_1mcos_tensor = get_theta_tensors(params, param_ids)
    ci_strings, strs2addr = get_ci_strings(n_qubits, n_elec, hcb, strs2addr=True)

    if init_state is None:
        civector = get_init_civector(len(ci_strings))
    else:
        civector = tc.backend.convert_to_tensor(init_state)

    for theta_sin, theta_1mcos, f_idx in zip(theta_sin_tensor, theta_1mcos_tensor, ex_ops):
        fket_permutation, fket_phase, f2ket_phase = get_operators(n_qubits, n_elec, strs2addr, f_idx, ci_strings, hcb)
        civector = evolve_excitation_nocache(
            civector, fket_permutation, fket_phase, f2ket_phase, theta_1mcos, theta_sin
        )

    return civector.reshape(-1)


def get_energy_and_grad_civector_nocache(
    params, hamiltonian, n_qubits, n_elec, ex_ops: Tuple, param_ids: Tuple, hcb: bool = False, init_state=None
):
    ket = get_civector_nocache(params, n_qubits, n_elec, ex_ops, param_ids, hcb, init_state)
    bra = apply_op(hamiltonian, ket)
    energy = bra @ ket

    gradients_beforesum = _get_gradients_civector_nocache(bra, ket, params, n_qubits, n_elec, ex_ops, param_ids, hcb)
    gradients_beforesum = tc.backend.numpy(gradients_beforesum)

    gradients = np.zeros(params.shape)
    for grad, param_id in zip(gradients_beforesum, param_ids):
        gradients[param_id] += grad

    return energy, 2 * gradients


@partial(jit, static_argnums=[3, 4, 5, 6, 7])
def _get_gradients_civector_nocache(bra, ket, params, n_qubits, n_elec, ex_ops, param_ids, hcb):
    ci_strings, strs2addr = get_ci_strings(n_qubits, n_elec, hcb, True)
    theta_sin_tensor, theta_1mcos_tensor = get_theta_tensors(params, param_ids)

    gradients_beforesum = []
    for theta_sin, theta_1mcos, f_idx in reversed(list(zip(theta_sin_tensor, theta_1mcos_tensor, ex_ops))):
        fket_permutation, fket_phase, f2ket_phase = get_operators(n_qubits, n_elec, strs2addr, f_idx, ci_strings, hcb)
        bra = evolve_excitation_nocache(bra, fket_permutation, fket_phase, f2ket_phase, theta_1mcos, -theta_sin)
        ket = evolve_excitation_nocache(ket, fket_permutation, fket_phase, f2ket_phase, theta_1mcos, -theta_sin)
        fket = ket[fket_permutation] * fket_phase
        grad = bra @ fket
        gradients_beforesum.append(grad)
    gradients_beforesum = list(reversed(gradients_beforesum))
    gradients_beforesum = tc.backend.convert_to_tensor(gradients_beforesum)

    return gradients_beforesum


def apply_excitation_civector(civector, n_qubits, n_elec, f_idx, hcb):
    _, fket_permutation, fket_phase, _ = get_operator_tensors(n_qubits, n_elec, ex_ops=(f_idx,), hcb=hcb)
    return civector[fket_permutation[0]] * fket_phase[0]


def apply_excitation_civector_nocache(civector, n_qubits, n_elec, f_idx, hcb):
    ci_strings, strs2addr = get_ci_strings(n_qubits, n_elec, hcb, strs2addr=True)
    fket_permutation, fket_phase, _ = get_operators(n_qubits, n_elec, strs2addr, f_idx, ci_strings, hcb)
    return civector[fket_permutation] * fket_phase



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


import logging
from functools import partial
from time import time
from typing import Callable, Union, Any, List, Tuple

import numpy as np
from scipy.optimize import minimize
from noisyopt import minimizeSPSA

from openfermion import (
    jordan_wigner,
    bravyi_kitaev,
    binary_code_transform,
    checksum_code,
    parity_code,
    get_sparse_operator,
    QubitOperator,
    FermionOperator,
)
from qiskit import QuantumCircuit
import tensorcircuit as tc
from tensorcircuit import Circuit, DMCircuit
from tensorcircuit.noisemodel import NoiseConf, circuit_with_noise

from tencirchem.static.engine_hea import (
    get_statevector,
    get_densitymatrix,
    get_energy_tensornetwork,
    get_energy_tensornetwork_noise,
    get_energy_tensornetwork_shot,
    get_energy_tensornetwork_noise_shot,
    get_energy_qpu,
    get_energy_and_grad_tensornetwork,
    get_energy_and_grad_tensornetwork_noise,
    get_energy_and_grad_tensornetwork_shot,
    get_energy_and_grad_tensornetwork_noise_shot,
    get_energy_and_grad_qpu,
)
from tencirchem.static.hamiltonian import get_hop_from_integral
from tencirchem.utils.misc import reverse_fop_idx, scipy_opt_wrap, reverse_qop_idx
from tencirchem.utils.circuit import get_circuit_dataframe

logger = logging.getLogger(__file__)


Tensor = Any


def get_noise_conf(probability=0.02):
    noise_conf = NoiseConf()
    # note TC definition of error probability
    channel = tc.channels.isotropicdepolarizingchannel(probability, 2)

    def condition(qir):
        return len(qir["index"]) == 2

    noise_conf.add_noise_by_condition(condition, channel)
    return noise_conf


def binary(fermion_operator: FermionOperator, n_modes: int, n_elec: int) -> QubitOperator:
    """
    Performs binary transformation.

    Parameters
    ----------
    fermion_operator: FermionOperator
        The fermion operator.
    n_modes: int
        The number of modes.
    n_elec: int
        The number of electrons.

    Returns
    -------
    qubit_operator: QubitOperator
    """
    return binary_code_transform(fermion_operator, 2 * checksum_code(n_modes // 2, (n_elec // 2) % 2))


def _parity(fermion_operator, n_modes):
    return binary_code_transform(fermion_operator, parity_code(n_modes))


def parity(fermion_operator: FermionOperator, n_modes: int, n_elec: int) -> QubitOperator:
    """
    Performs parity transformation.

    Parameters
    ----------
    fermion_operator: FermionOperator
        The fermion operator.
    n_modes: int
        The number of modes (spin-orbitals).
    n_elec: int
        The number of electrons.

    Returns
    -------
    qubit_operator: QubitOperator
    """
    qubit_operator = _parity(reverse_fop_idx(fermion_operator, n_modes), n_modes)
    res = 0
    assert n_modes % 2 == 0
    reduction_indices = [n_modes // 2 - 1, n_modes - 1]
    phase_alpha = (-1) ** (n_elec // 2)
    for qop in qubit_operator:
        # qop example: 0.5 [Z1 X2 X3]
        pauli_string, coeff = next(iter(qop.terms.items()))
        # pauli_string example: ((1, 'Z'), (2, 'X'), (3, 'X'))
        # coeff example: 0.5
        new_pauli_string = []
        for idx, symbol in pauli_string:
            is_alpha = idx <= reduction_indices[0]
            if idx in reduction_indices:
                if symbol in ["X", "Y"]:
                    # discard this term because the bit will never change
                    continue
                else:
                    assert symbol == "Z"
                    if is_alpha:
                        coeff *= phase_alpha
                    continue
            if not is_alpha:
                idx -= 1
            new_pauli_string.append((idx, symbol))
        qop.terms = {tuple(new_pauli_string): coeff}
        res += qop
    return res


def get_ry_circuit(params: Tensor, n_qubits: int, n_layers: int) -> Circuit:
    """
    Get the parameterized :math:`R_y` circuit.

    Parameters
    ----------
    params: Tensor
        The circuit parameters.
    n_qubits: int
        The number of qubits.
    n_layers: int
        The number of layers in the ansatz.

    Returns
    -------
    c: Circuit
    """
    c = Circuit(n_qubits)
    params = params.reshape(n_layers + 1, n_qubits)
    for i in range(n_qubits):
        c.ry(i, theta=params[0, i])
    c.barrier_instruction(*range(n_qubits))
    for l in range(n_layers):
        for i in range(n_qubits - 1):
            c.cnot(i, (i + 1))
        for i in range(n_qubits):
            c.ry(i, theta=params[l + 1, i])
        c.barrier_instruction(*range(n_qubits))
    return c


class HEA:
    """
    Run hardware-efficient ansatz calculation.
    For a comprehensive tutorial see :doc:`/tutorial_jupyter/noisy_simulation`.
    """

    @classmethod
    def ry(
        cls,
        int1e: np.ndarray,
        int2e: np.ndarray,
        n_elec: int,
        e_core: float,
        n_layers: int,
        init_circuit: Circuit = None,
        mapping: str = "parity",
        **kwargs,
    ):
        """
        Construct the HEA class from electron integrals and custom quantum circuit.
        Overlap integral is assumed to be identity.
        Parity transformation is used to transform from fermion operators to qubit operators.

        Parameters
        ----------
        int1e: np.ndarray
            One-body integral in spatial orbital.
        int2e: np.ndarray
            Two-body integral, in spatial orbital, chemists' notation, and without considering symmetry.
        n_elec: int
            The number of electrons
        e_core: float
            The nuclear energy or core energy if active space approximation is involved.
        n_layers: int
            The number of layers in the ansatz.
        init_circuit: Circuit
            The initial circuit before the :math:`R_y` ansatz. Defaults to None,
            which creates an HF initial state.
        mapping: str
            The fermion to qubit mapping. Supported mappings are ``"parity"``,
            and ``"bravyi-kitaev"``.

        kwargs:
            Other arguments to be passed to the :func:`__init__` function such as ``engine``.

        Returns
        -------
        hea: :class:`HEA`
             An HEA instance
        """
        n_sorb = 2 * len(int1e)
        if mapping == "parity":
            n_qubits = n_sorb - 2
        elif mapping in ["jordan-wigner", "bravyi-kitaev"]:
            n_qubits = n_sorb
        else:
            raise ValueError(f"Unknown mapping: {mapping}")

        init_guess = np.random.random((n_layers + 1, n_qubits)).ravel()

        def get_circuit(params):
            if init_circuit is None:
                c = Circuit(n_qubits)
            else:
                c = Circuit.from_qir(init_circuit.to_qir(), init_circuit.circuit_param)
            return c.append(get_ry_circuit(params, n_qubits, n_layers))

        return cls.from_integral(int1e, int2e, n_elec, e_core, mapping, get_circuit, init_guess, **kwargs)

    @classmethod
    def from_integral(
        cls,
        int1e: np.ndarray,
        int2e: np.ndarray,
        n_elec: int,
        e_core: float,
        mapping: str,
        circuit: Union[Callable, QuantumCircuit],
        init_guess: np.ndarray,
        **kwargs,
    ):
        """
        Construct the HEA class from electron integrals and custom quantum circuit.
        Overlap integral is assumed to be identity.

        Parameters
        ----------
        int1e: np.ndarray
            One-body integral in spatial orbital.
        int2e: np.ndarray
            Two-body integral, in spatial orbital, chemists' notation, and without considering symmetry.
        n_elec: int
            The number of electrons
        e_core: float
            The nuclear energy or core energy if active space approximation is involved.
        mapping: str
            The fermion to qubit mapping. Supported mappings are ``"parity"``,
            and ``"bravyi-kitaev"``.
        circuit: Callable or QuantumCircuit
            The ansatz as a function or Qiskit :class:`QuantumCircuit`
        init_guess: list of float or :class:`np.ndarray`
            The parameter initial guess.

        kwargs:
            Other arguments to be passed to the :func:`__init__` function such as ``engine``.

        Returns
        -------
        hea: :class:`HEA`
             An HEA instance
        """
        hop = get_hop_from_integral(int1e, int2e) + e_core
        n_sorb = 2 * len(int1e)
        if mapping == "parity":
            h_qubit_op = parity(hop, n_sorb, n_elec)
        elif mapping == "jordan-wigner":
            h_qubit_op = reverse_qop_idx(jordan_wigner(hop), n_sorb)
        elif mapping == "bravyi-kitaev":
            h_qubit_op = reverse_qop_idx(bravyi_kitaev(hop, n_sorb), n_sorb)
        else:
            raise ValueError(f"Unknown mapping: {mapping}")

        instance = cls(h_qubit_op, circuit, init_guess, **kwargs)

        instance.int1e = int1e
        instance.int2e = int2e
        instance.e_core = e_core
        instance.hop = hop
        return instance

    def __init__(
        self,
        h_qubit_op: QubitOperator,
        circuit: Union[Callable, QuantumCircuit],
        init_guess: Union[List[float], np.ndarray],
        engine: str = None,
        engine_conf: NoiseConf = None,
    ):
        """
        Construct the HEA class from Hamiltonian in :class:`QubitOperator` form and the ansatz.

        Parameters
        ----------
        h_qubit_op: QubitOperator
            Hamiltonian as openfermion :class:`QubitOperator`.
        circuit: Callable or QuantumCircuit
            The ansatz as a function or Qiskit :class:`QuantumCircuit`
        init_guess: list of float or :class:`np.ndarray`
            The parameter initial guess.
        engine: str, optional
            The engine to run the calculation. See :ref:`advanced:Engines` for details.
            Defaults to ``"tensornetwork"``.
        engine_conf: NoiseConf, optional
            The noise configuration for the circuit. Defaults to None, in which case
            if a noisy engine is used, an isotropic depolarizing error channel for all 2-qubit gates
            with :math:`p=0.02` is added to the circuit.
        """
        self._check_engine(engine)

        if engine is None:
            engine = "tensornetwork"
        if engine == "tensornetwork" and engine_conf is not None:
            raise ValueError("Tensornetwork engine does not have engine configuration")
        if engine.startswith("tensornetwork-noise") and engine_conf is None:
            engine_conf = get_noise_conf()

        init_guess = np.array(init_guess)

        self.h_qubit_op = h_qubit_op
        self.h_array = np.array(get_sparse_operator(self.h_qubit_op).todense())

        if isinstance(circuit, Callable):
            self.get_circuit = circuit
            # sanity check
            c: Circuit = self.get_circuit(init_guess)
            if isinstance(c, DMCircuit):
                raise TypeError("`circuit` function should return Circuit instead of DMCircuit")
            self.n_qubits = c.circuit_param["nqubits"]
        elif isinstance(circuit, QuantumCircuit):

            def get_circuit(params):
                return Circuit.from_qiskit(circuit, binding_params=params)

            self.get_circuit = get_circuit
            assert circuit.num_parameters == len(init_guess)
            self.n_qubits = circuit.num_qubits
        else:
            raise TypeError("circuit must be callable or qiskit QuantumCircuit")

        if init_guess.ndim != 1:
            raise ValueError(f"Init guess should be one-dimensional. Got shape {init_guess}")
        self.init_guess = init_guess
        self.engine = engine
        self.engine_conf = engine_conf
        self.shots = 4096
        self._grad = "param-shift"

        self.scipy_minimize_options = None
        self._params = None
        self.opt_res = None

    def get_dmcircuit(self, params: Tensor = None, noise_conf: NoiseConf = None) -> DMCircuit:
        """
        Get the :class:`DMCircuit` with noise.
        Only valid for ``"tensornetwork-noise"`` and ``"tensornetwork-noise&shot"`` engines.

        Parameters
        ----------
        params: Tensor, optional
            The circuit parameters. Defaults to None, which uses the optimized parameter
            and :func:`kernel` must be called before.
        noise_conf: NoiseConf, optional
            The noise configuration for the circuit. Defaults to None, in which case
            ``self.engine_conf`` is used.

        Returns
        -------
        dmcircuit: DMCircuit
        """
        params = self._check_params_argument(params)
        dmcircuit = self.get_dmcircuit_no_noise(params)
        if noise_conf is None:
            assert self.engine.startswith("tensornetwork-noise")
            noise_conf = self.engine_conf
        dmcircuit = circuit_with_noise(dmcircuit, noise_conf)
        return dmcircuit

    def get_dmcircuit_no_noise(self, params: Tensor = None) -> DMCircuit:
        """
        Get the :class:`DMCircuit` without noise.

        Parameters
        ----------
        params: Tensor, optional
            The circuit parameters. Defaults to None, which uses the optimized parameter
            and :func:`kernel` must be called before.

        Returns
        -------
        dmcircuit: DMCircuit
        """
        qir = self.get_circuit(params).to_qir()
        dmcircuit = DMCircuit.from_qir(qir)
        return dmcircuit

    def _check_engine(self, engine):
        supported_engine = [
            None,
            "tensornetwork",
            "tensornetwork-noise",
            "tensornetwork-shot",
            "tensornetwork-noise&shot",
            "qpu",
        ]
        if not engine in supported_engine:
            raise ValueError(f"Engine '{engine}' not supported")

    def _check_params_argument(self, params, strict=True):
        if params is None:
            if self.params is not None:
                params = self.params
            else:
                if strict:
                    raise ValueError("Run the `.kernel` method to determine the parameters first")
                else:
                    if self.init_guess is not None:
                        params = self.init_guess
                    else:
                        params = np.zeros(self.n_params)

        if len(params) != self.n_params:
            raise ValueError(f"Incompatible parameter shape. {self.n_params} is desired. Got {len(params)}")
        return tc.backend.convert_to_tensor(params).astype(tc.rdtypestr)

    def statevector(self, params: Tensor = None) -> Tensor:
        """
        Evaluate the circuit state vector without considering noise.
        Valid for ``"tensornetwork"`` and ``"tensornetwork-shot"`` engine.

        Parameters
        ----------
        params: Tensor, optional
            The circuit parameters. Defaults to None, which uses the optimized parameter
            and :func:`kernel` must be called before.

        Returns
        -------
        statevector: Tensor
            Corresponding state vector

        See Also
        --------
        densitymatrix: Evaluate the circuit density matrix in the presence of circuit noise.
        energy: Evaluate the total energy.
        energy_and_grad: Evaluate the total energy and parameter gradients.

        Examples
        --------
        >>> import numpy as np
        >>> from tencirchem import UCC, HEA
        >>> from tencirchem.molecule import h2
        >>> ucc = UCC(h2)
        >>> hea = HEA.ry(ucc.int1e, ucc.int2e, ucc.n_elec, ucc.e_core, n_layers=1)
        >>> np.round(hea.statevector([0, np.pi, 0, 0]), 8)  # HF state
        array([0.+0.j, 1.+0.j, 0.+0.j, 0.+0.j])
        """
        # Evaluate the circuit state vector without noise.
        # only engine is "tensornetwork"
        params = self._check_params_argument(params)
        statevector = get_statevector(params, self.get_circuit)
        return statevector

    def densitymatrix(self, params: Tensor = None) -> Tensor:
        """
        Evaluate the circuit density matrix in the presence of circuit noise.
        Only valid for ``"tensornetwork-noise"`` and ``"tensornetwork-noise&shot"`` engines.

        Parameters
        ----------
        params: Tensor, optional
            The circuit parameters. Defaults to None, which uses the optimized parameter
            and :func:`kernel` must be called before.

        Returns
        -------
        densitymatrix: Tensor

        See Also
        --------
        statevector: Evaluate the circuit state vector.
        energy: Evaluate the total energy.
        energy_and_grad: Evaluate the total energy and parameter gradients.

        Examples
        --------
        >>> import numpy as np
        >>> from tencirchem import UCC, HEA
        >>> from tencirchem.molecule import h2
        >>> ucc = UCC(h2)
        >>> hea = HEA.ry(ucc.int1e, ucc.int2e, ucc.n_elec, ucc.e_core,
        ...         n_layers=1, engine="tensornetwork-noise")
        >>> np.round(hea.densitymatrix([0, np.pi, 0, 0]), 8)  # HF state with noise
        array([[0.00533333+0.j, 0.        +0.j, 0.        +0.j, 0.        +0.j],
               [0.        +0.j, 0.984     +0.j, 0.        +0.j, 0.        +0.j],
               [0.        +0.j, 0.        +0.j, 0.00533333+0.j, 0.        +0.j],
               [0.        +0.j, 0.        +0.j, 0.        +0.j, 0.00533333+0.j]])
        """
        # engines are "tensornetwork-noise" and "tensornetwork-noise&shot"
        params = self._check_params_argument(params)
        # the last two arguments should be identical and not garbage collected for each call for `jit` to work
        densitymatrix = get_densitymatrix(params, self.get_dmcircuit_no_noise, self.engine_conf)
        return densitymatrix

    def energy(self, params: Tensor = None, engine: str = None) -> float:
        """
        Evaluate the total energy.

        Parameters
        ----------
        params: Tensor, optional
            The circuit parameters. Defaults to None, which uses the optimized parameter
            and :func:`kernel` must be called before.
        engine: str, optional
            The engine to use. Defaults to ``None``, which uses ``self.engine``.

        Returns
        -------
        energy: float
            Total energy

        See Also
        --------
        statevector: Evaluate the circuit state vector.
        densitymatrix: Evaluate the circuit density matrix in the presence of circuit noise.
        energy_and_grad: Evaluate the total energy and parameter gradients.

        Examples
        --------
        >>> import numpy as np
        >>> from tencirchem import UCC, HEA
        >>> from tencirchem.molecule import h2
        >>> ucc = UCC(h2)
        >>> hea = HEA.ry(ucc.int1e, ucc.int2e, ucc.n_elec, ucc.e_core,
        ...         n_layers=1, engine="tensornetwork-noise")
        >>> # HF state, no noise
        >>> round(hea.energy([0, np.pi, 0, 0], "tensornetwork"), 8)
        -1.11670614
        >>> # HF state, gate noise
        >>> round(hea.energy([0, np.pi, 0, 0], "tensornetwork-noise"), 4)
        -1.1001
        >>> # HF state, measurement noise. Set the number of shots by `hea.shots`
        >>> round(hea.energy([0, np.pi, 0, 0], "tensornetwork-shot"), 1)
        -1.1
        >>> # HF state, gate+measurement noise
        >>> hea.energy([0, np.pi, 0, 0], "tensornetwork-noise&shot")  # doctest:+ELLIPSIS
        -1...
        """
        params = self._check_params_argument(params)
        if engine is None:
            engine = self.engine
        if engine == "tensornetwork":
            e = get_energy_tensornetwork(params, self.h_array, self.get_circuit)
        elif engine == "tensornetwork-noise":
            e = get_energy_tensornetwork_noise(params, self.h_array, self.get_dmcircuit_no_noise, self.engine_conf)
        elif engine == "tensornetwork-shot":
            e = get_energy_tensornetwork_shot(
                params,
                tuple(self.h_qubit_op.terms.keys()),
                list(self.h_qubit_op.terms.values()),
                self.get_circuit,
                self.shots,
            )
        elif engine == "tensornetwork-noise&shot":
            e = get_energy_tensornetwork_noise_shot(
                params,
                tuple(self.h_qubit_op.terms.keys()),
                list(self.h_qubit_op.terms.values()),
                self.get_dmcircuit_no_noise,
                self.engine_conf,
                self.shots,
            )
        else:
            assert engine == "qpu"
            e = get_energy_qpu(
                params,
                tuple(self.h_qubit_op.terms.keys()),
                list(self.h_qubit_op.terms.values()),
                self.get_circuit,
                self.shots,
            )
        return e

    def energy_and_grad(self, params: Tensor = None, engine: str = None, grad: str = None) -> Tuple[float, Tensor]:
        """
        Evaluate the total energy and parameter gradients using parameter-shift rule.

        Parameters
        ----------
        params: Tensor, optional
            The circuit parameters. Defaults to None, which uses the optimized parameter
            and :func:`kernel` must be called before.
        engine: str, optional
            The engine to use. Defaults to ``None``, which uses ``self.engine``.
        grad: str, optional
            The algorithm to use for the gradient. Defaults to ``None``, which means ``self.grad`` will be used.
            Possible options are ``"param-shift"`` for parameter-shift rule and
            ``"autodiff"`` for auto-differentiation.
            Note that ``"autodiff"`` is not compatible with ``"tensornetwork-shot"``
            and ``"tensornetwork-noise&shot"`` engine.

        Returns
        -------
        energy: float
            Total energy
        grad: Tensor
            The parameter gradients

        See Also
        --------
        statevector: Evaluate the circuit state vector.
        densitymatrix: Evaluate the circuit density matrix in the presence of circuit noise.
        energy: Evaluate the total energy.
        """

        params = self._check_params_argument(params)

        if engine is None:
            engine = self.engine

        if grad is None:
            grad = self.grad

        if grad == "free":
            raise ValueError("Must provide a gradient algorithm")

        if engine == "tensornetwork":
            e, grad_array = get_energy_and_grad_tensornetwork(params, self.h_array, self.get_circuit, grad)
        elif engine == "tensornetwork-noise":
            e, grad_array = get_energy_and_grad_tensornetwork_noise(
                params, self.h_array, self.get_dmcircuit_no_noise, self.engine_conf, grad
            )
        elif engine == "tensornetwork-shot":
            if grad == "autodiff":
                raise ValueError(f"Engine {engine} is incompatible with grad method {grad}")
            e, grad_array = get_energy_and_grad_tensornetwork_shot(
                params,
                tuple(self.h_qubit_op.terms.keys()),
                list(self.h_qubit_op.terms.values()),
                self.get_circuit,
                self.shots,
                grad,
            )
        elif engine == "tensornetwork-noise&shot":
            if grad == "autodiff":
                raise ValueError(f"Engine {engine} is incompatible with grad method {grad}")
            e, grad_array = get_energy_and_grad_tensornetwork_noise_shot(
                params,
                tuple(self.h_qubit_op.terms.keys()),
                list(self.h_qubit_op.terms.values()),
                self.get_dmcircuit_no_noise,
                self.engine_conf,
                self.shots,
                grad,
            )
        else:
            assert engine == "qpu"
            if grad == "autodiff":
                raise ValueError(f"Engine {engine} is incompatible with grad method {grad}")
            e, grad_array = get_energy_and_grad_qpu(
                params,
                tuple(self.h_qubit_op.terms.keys()),
                list(self.h_qubit_op.terms.values()),
                self.get_circuit,
                self.shots,
                grad,
            )
        return e, grad_array

    def kernel(self):
        logger.info("Begin optimization")

        func, stating_time = self.get_opt_function(with_time=True)

        time1 = time()
        if self.grad == "free":
            if self.engine in ["tensornetwork", "tensornetwork-noise", "qpu"]:
                opt_res = minimize(func, x0=self.init_guess, method="COBYLA", options=self.scipy_minimize_options)
            else:
                assert self.engine in ["tensornetwork-shot", "tensornetwork-noise&shot"]
                opt_res = minimizeSPSA(func, x0=self.init_guess, paired=False, niter=100, disp=True)
        else:
            opt_res = minimize(
                func, x0=self.init_guess, jac=True, method="L-BFGS-B", options=self.scipy_minimize_options
            )

        time2 = time()

        if not opt_res.success:
            logger.warning("Optimization failed. See `.opt_res` for details.")

        opt_res["staging_time"] = stating_time
        opt_res["opt_time"] = time2 - time1
        opt_res["init_guess"] = self.init_guess
        opt_res["e"] = float(opt_res.fun)
        self.opt_res = opt_res
        # prepare for future modification
        self.params = opt_res.x.copy()
        return opt_res.e

    def get_opt_function(self, grad: str = None, with_time: bool = False) -> Union[Callable, Tuple[Callable, float]]:
        """
        Returns the cost function in SciPy format for optimization.
        Basically a wrapper to :func:`energy_and_grad` or :func:`energy`,


        Parameters
        ----------
        with_time: bool, optional
            Whether return staging time. Defaults to False.
        grad: str, optional
            The algorithm to use for the gradient. Defaults to ``None``, which means ``self.grad`` will be used.
            Possible options are ``"param-shift"`` for parameter-shift rule and
            ``"autodiff"`` for auto-differentiation.
            Note that ``"autodiff"`` is not compatible with ``"tensornetwork-noise&shot"`` engine.
        Returns
        -------
        opt_function: Callable
            The optimization cost function in SciPy format.
        time: float
            Staging time. Returned when ``with_time`` is set to ``True``.
        """
        if grad is None:
            grad = self.grad

        if grad != "free":
            func = scipy_opt_wrap(partial(self.energy_and_grad, engine=self.engine))
        else:
            func = scipy_opt_wrap(partial(self.energy, engine=self.engine), gradient=False)

        time1 = time()
        if tc.backend.name == "jax":
            logger.info("JIT compiling the circuit")
            _ = func(np.zeros(self.n_params))
            logger.info("Circuit JIT compiled")
        time2 = time()
        if with_time:
            return func, time2 - time1
        return func

    def print_circuit(self):
        c = self.get_circuit(self.init_guess)
        df = get_circuit_dataframe(c)

        def format_flop(f):
            return f"{f:.3e}"

        formatters = {"flop": format_flop}
        print(df.to_string(index=True, formatters=formatters))

    def print_summary(self):
        print("############################### Circuit ###############################")
        self.print_circuit()
        print("######################### Optimization Result #########################")
        if self.opt_res is None:
            print("Optimization not run yet")
        else:
            print(self.opt_res)

    @property
    def grad(self):
        return self._grad

    @grad.setter
    def grad(self, v):
        if not v in ["param-shift", "autodiff", "free"]:
            raise ValueError(f"Invalid gradient method {v}")
        self._grad = v

    @property
    def n_params(self):
        """The number of parameter in the ansatz/circuit."""
        return len(self.init_guess)

    @property
    def params(self):
        """The circuit parameters."""
        if self._params is not None:
            return self._params
        if self.opt_res is not None:
            return self.opt_res.x
        return None

    @params.setter
    def params(self, params):
        self._params = params



import numpy as np

from tencirchem import clear_cache, UCCSD
from tencirchem.molecule import h4


def test_clear_cache():
    uccsd = UCCSD(h4)
    e1 = uccsd.kernel()
    clear_cache()
    e2 = uccsd.kernel()
    np.testing.assert_allclose(e2, e1)



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from time import time
import logging
from typing import Tuple, List

import numpy as np
from pyscf.gto.mole import Mole

from tencirchem.static.ucc import UCC

logger = logging.getLogger(__name__)


class KUPCCGSD(UCC):
    """
    Run :math:`k`-UpCCGSD calculation.
    The interfaces are similar to :class:`UCCSD <tencirchem.UCCSD>`.
    """

    def __init__(
        self,
        mol: Mole,
        active_space: Tuple[int, int] = None,
        mo_coeff: np.ndarray = None,
        k: int = 3,
        n_tries: int = 1,
        engine: str = None,
        run_hf: bool = True,
        run_mp2: bool = True,
        run_ccsd: bool = True,
        run_fci: bool = True,
    ):
        r"""
        Initialize the class with molecular input.

        Parameters
        ----------
        mol: Mole
            The molecule as PySCF ``Mole`` object.
        active_space: Tuple[int, int], optional
            Active space approximation. The first integer is the number of electrons and the second integer is
            the number or spatial-orbitals. Defaults to None.
        mo_coeff: np.ndarray, optional
            Molecule coefficients. If provided then RHF is skipped.
            Can be used in combination with the ``init_state`` attribute.
            Defaults to None which means RHF orbitals are used.
        k: int, optional
            The number of layers in the ansatz. Defaults to 3
        n_tries: int, optional
            The number of different initial points used for VQE calculation.
            For large circuits usually a lot of runs are required for good accuracy.
            Defaults to 1.
        engine: str, optional
            The engine to run the calculation. See :ref:`advanced:Engines` for details.
        run_hf: bool, optional
            Whether run HF for molecule orbitals. Defaults to ``True``.
        run_mp2: bool, optional
            Whether run MP2 for energy reference. Defaults to ``True``.
        run_ccsd: bool, optional
            Whether run CCSD for energy reference. Defaults to ``True``.
        run_fci: bool, optional
            Whether run FCI for energy reference. Defaults to ``True``.

        See Also
        --------
        tencirchem.UCCSD
        tencirchem.PUCCD
        tencirchem.UCC
        """
        super().__init__(
            mol,
            init_method=None,
            active_space=active_space,
            mo_coeff=mo_coeff,
            engine=engine,
            run_hf=run_hf,
            run_mp2=run_mp2,
            run_ccsd=run_ccsd,
            run_fci=run_fci,
        )
        # the number of layers
        self.k = k
        # the number of different initialization
        self.n_tries = n_tries
        self.ex_ops, self.param_ids, self.init_guess = self.get_ex_ops(self.t1, self.t2)
        self.init_guess_list = [self.init_guess]
        for _ in range(self.n_tries - 1):
            self.init_guess_list.append(np.random.rand(self.n_params) - 0.5)
        self.e_tries_list = []
        self.opt_res_list = []
        self.staging_time = self.opt_time = None

    def kernel(self):
        _, stating_time = self.get_opt_function(with_time=True)

        time1 = time()
        for i in range(self.n_tries):
            logger.info(f"k-UpCCGSD try {i}")
            if self.n_tries == 1:
                if not np.allclose(self.init_guess, self.init_guess_list[0]):
                    logger.info("Inconsistent `self.init_guess` and `self.init_guess_list`.  Use `self.init_guess`.")
            else:
                self.init_guess = self.init_guess_list[i]
            super().kernel()
            logger.info(f"k-UpCCGSD try {i} energy {self.opt_res.fun}")
            self.opt_res_list.append(self.opt_res)
        self.opt_res_list.sort(key=lambda x: x.fun)
        self.e_tries_list = [float(res.fun) for res in self.opt_res_list]
        time2 = time()

        self.staging_time = stating_time
        self.opt_time = time2 - time1
        self.opt_res = self.opt_res_list[0]
        self.opt_res.e_tries = self.e_tries_list

        if not self.opt_res.success:
            logger.warning("Optimization failed. See `.opt_res` for details.")

        self.init_guess = self.opt_res.init_guess
        return float(self.opt_res.e)

    def get_ex_ops(self, t1: np.ndarray = None, t2: np.ndarray = None) -> Tuple[List[Tuple], List[int], np.ndarray]:
        """
        Get one-body and two-body excitation operators for :math:`k`-UpCCGSD ansatz.
        The excitations are generalized and two-body excitations are restricted to paired ones.
        Initial guesses are generated randomly.

        Parameters
        ----------
        t1: np.ndarray, optional
            Not used. Kept for consistency with the parent method.
        t2: np.ndarray, optional
            Not used. Kept for consistency with the parent method.

        Returns
        -------
        ex_op: List[Tuple]
            The excitation operators. Each operator is represented by a tuple of ints.
        param_ids: List[int]
            The mapping from excitations to parameters.
        init_guess: np.ndarray
            The initial guess for the parameters.

        See Also
        --------
        get_ex1_ops: Get generalized one-body excitation operators.
        get_ex2_ops: Get generalized paired two-body excitation operators.

        Examples
        --------
        >>> from tencirchem import KUPCCGSD
        >>> from tencirchem.molecule import h2
        >>> kupccgsd = KUPCCGSD(h2)
        >>> ex_op, param_ids, init_guess = kupccgsd.get_ex_ops()
        >>> ex_op
        [(1, 3, 2, 0), (3, 2), (1, 0), (1, 3, 2, 0), (3, 2), (1, 0), (1, 3, 2, 0), (3, 2), (1, 0)]
        >>> param_ids
        [0, 1, 1, 2, 3, 3, 4, 5, 5]
        >>> init_guess  # doctest:+ELLIPSIS
        array([...])
        """
        ex1_ops, ex1_param_id, _ = self.get_ex1_ops()
        ex2_ops, ex2_param_id, _ = self.get_ex2_ops()

        ex_op = []
        param_ids = [-1]
        for _ in range(self.k):
            ex_op.extend(ex2_ops + ex1_ops)
            param_ids.extend([i + param_ids[-1] + 1 for i in ex2_param_id])
            param_ids.extend([i + param_ids[-1] + 1 for i in ex1_param_id])
        init_guess = np.random.rand(max(param_ids) + 1) - 0.5
        return ex_op, param_ids[1:], init_guess

    def get_ex1_ops(self, t1: np.ndarray = None) -> Tuple[List[Tuple], List[int], np.ndarray]:
        """
        Get generalized one-body excitation operators.

        Parameters
        ----------
        t1: np.ndarray, optional
            Not used. Kept for consistency with the parent method.

        Returns
        -------
        ex_op: List[Tuple]
            The excitation operators. Each operator is represented by a tuple of ints.
        param_ids: List[int]
            The mapping from excitations to parameters.
        init_guess: np.ndarray
            The initial guess for the parameters.

        See Also
        --------
        get_ex2_ops: Get generalized paired two-body excitation operators.
        get_ex_ops: Get one-body and two-body excitation operators for :math:`k`-UpCCGSD ansatz.
        """
        assert t1 is None
        no, nv = self.no, self.nv

        ex1_ops = []
        ex1_param_id = [-1]

        for a in range(no + nv):
            for i in range(a):
                # alpha to alpha
                ex_op_a = (no + nv + a, no + nv + i)
                # beta to beta
                ex_op_b = (a, i)
                ex1_ops.extend([ex_op_a, ex_op_b])
                ex1_param_id.extend([ex1_param_id[-1] + 1] * 2)

        ex1_init_guess = np.zeros(max(ex1_param_id) + 1)
        return ex1_ops, ex1_param_id[1:], ex1_init_guess

    def get_ex2_ops(self, t2: np.ndarray = None) -> Tuple[List[Tuple], List[int], np.ndarray]:
        """
        Get generalized paired two-body excitation operators.

        Parameters
        ----------
        t2: np.ndarray, optional
            Not used. Kept for consistency with the parent method.

        Returns
        -------
        ex_op: List[Tuple]
            The excitation operators. Each operator is represented by a tuple of ints.
        param_ids: List[int]
            The mapping from excitations to parameters.
        init_guess: np.ndarray
            The initial guess for the parameters.

        See Also
        --------
        get_ex1_ops: Get one-body excitation operators.
        get_ex_ops: Get one-body and two-body excitation operators for :math:`k`-UpCCGSD ansatz.
        """

        assert t2 is None
        no, nv = self.no, self.nv

        ex2_ops = []
        ex2_param_id = [-1]

        for a in range(no + nv):
            for i in range(a):
                # i correspond to a and j correspond to b, as in PySCF convention
                # otherwise the t2 amplitude has incorrect phase
                # paired
                ex_op_ab = (a, no + nv + a, no + nv + i, i)
                ex2_ops.append(ex_op_ab)
                ex2_param_id.append(ex2_param_id[-1] + 1)

        ex2_init_guess = np.zeros(max(ex2_param_id) + 1)
        return ex2_ops, ex2_param_id[1:], ex2_init_guess

    @property
    def e_kupccgsd(self):
        """
        Returns :math:`k`-UpCCGSD energy
        """
        return self.energy()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="tencirchem",
    version="2023.05",
    author="TenCirChem Authors",
    author_email="liw31@163.com",
    description="Efficient quantum computational chemistry based on TensorCircuit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tencent-quantum-lab/tencirchem",
    packages=setuptools.find_packages(),
    python_requires=">=3.7",
    include_package_data=True,
    install_requires=[
        "numpy",
        "scipy",
        "pandas",
        "tensorcircuit",
        "pyscf",
        "openfermion",
        "qiskit",
        "pylatexenc",
        "noisyopt",
        "renormalizer",
    ],
    extras_require={
        "jax": ["jax", "jaxlib"],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
)



import numpy as np
import pytest

from tencirchem import set_backend, PUCCD, UCC
from tencirchem.molecule import h4, h5p
from tests.static.test_static import get_random_integral_and_fci


def test_get_circuit():
    puccd = PUCCD(h5p)
    params = np.random.rand(puccd.n_params)
    s1 = puccd.get_circuit(params).state()
    s2 = puccd.get_circuit(params, trotter=True).state()
    np.testing.assert_allclose(s2, s1, atol=1e-10)
    s3 = puccd.get_circuit(params, givens_swap=True).state()
    np.testing.assert_allclose(s3, s1, atol=1e-10)


@pytest.mark.parametrize("hamiltonian", ["H4", "random integral"])
@pytest.mark.parametrize("engine", ["tensornetwork", "statevector", "civector", "civector-large"])
def test_puccd(hamiltonian, engine, reset_backend):
    if engine in ["tensornetwork", "statevector"]:
        set_backend("jax")

    if hamiltonian == "H4":
        puccd = PUCCD(h4, engine=engine)
        ucc = UCC(h4)
    else:
        nao = 4
        n_elec = 4
        int1e, int2e, _ = get_random_integral_and_fci(nao)
        puccd = PUCCD.from_integral(int1e, int2e, n_elec, engine=engine)
        ucc = UCC.from_integral(int1e, int2e, n_elec)

    # note the order
    # b^\dagger_i = a^\dagger_(nao + i) a^\dagger_(i)
    # b_j = a_(j) a_(nao + j)
    ucc.ex_ops = (
        (7, 3, 0, 4),
        (6, 2, 0, 4),
        (7, 3, 1, 5),
        (6, 2, 1, 5),
    )

    e1 = puccd.kernel()
    ucc.params = puccd.params
    e2 = ucc.energy()

    np.testing.assert_allclose(e1, e2, atol=1e-6)

    rdm1_puccd = puccd.make_rdm1()
    rdm1_ucc = ucc.make_rdm1()
    # slight difference in optimization minimum
    np.testing.assert_allclose(rdm1_puccd, rdm1_ucc, atol=1e-6)

    rdm2_puccd = puccd.make_rdm2(basis="MO")
    rdm2_ucc = ucc.make_rdm2(basis="MO")
    np.testing.assert_allclose(rdm2_puccd, rdm2_ucc, atol=1e-5)



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from functools import partial
from typing import Tuple
import logging

import tensorcircuit as tc

from tencirchem.utils.backend import jit, value_and_grad
from tencirchem.static.hamiltonian import apply_op
from tencirchem.static.ci_utils import get_ci_strings, civector_to_statevector, statevector_to_civector
from tencirchem.static.evolve_civector import (
    get_civector_nocache,
    get_energy_and_grad_civector,
    get_energy_and_grad_civector_nocache,
    apply_excitation_civector,
    apply_excitation_civector_nocache,
)
from tencirchem.static.evolve_civector import get_civector as get_civector_
from tencirchem.static.evolve_statevector import apply_excitation_statevector
from tencirchem.static.evolve_statevector import get_statevector as get_statevector_
from tencirchem.static.evolve_tensornetwork import get_statevector_tensornetwork
from tencirchem.static.evolve_pyscf import get_civector_pyscf, get_energy_and_grad_pyscf, apply_excitation_pyscf


logger = logging.getLogger(__name__)


GETVECTOR_MAP = {
    "tensornetwork": get_statevector_tensornetwork,
    "statevector": get_statevector_,
    "civector": get_civector_,
    "civector-large": get_civector_nocache,
    "pyscf": get_civector_pyscf,
}


def get_ket(params, n_qubits, n_elec, ex_ops, param_ids, hcb, init_state, ci_strings, engine):
    if param_ids is None:
        param_ids = range(len(ex_ops))
    func = GETVECTOR_MAP[engine]
    init_state = translate_init_state(init_state, n_qubits, ci_strings)
    ket = func(
        params,
        n_qubits=n_qubits,
        n_elec=n_elec,
        ex_ops=tuple(ex_ops),
        param_ids=tuple(param_ids),
        hcb=hcb,
        init_state=init_state,
    )
    return ket


def get_civector(params, n_qubits, n_elec, ex_ops, param_ids, hcb, init_state, engine):
    ci_strings = get_ci_strings(n_qubits, n_elec, hcb)
    ket = get_ket(params, n_qubits, n_elec, ex_ops, param_ids, hcb, init_state, ci_strings, engine)
    if engine.startswith("civector") or engine == "pyscf":
        civector = ket
    else:
        civector = ket[ci_strings]
    return civector


def get_statevector(params, n_qubits, n_elec, ex_ops, param_ids, hcb, init_state, engine):
    ci_strings = get_ci_strings(n_qubits, n_elec, hcb)
    ket = get_ket(params, n_qubits, n_elec, ex_ops, param_ids, hcb, init_state, ci_strings, engine)
    if engine.startswith("civector") or engine == "pyscf":
        statevector = civector_to_statevector(ket, n_qubits, ci_strings)
    else:
        statevector = ket
    return statevector


def get_energy(params, hamiltonian, n_qubits, n_elec, ex_ops: Tuple, param_ids: Tuple, hcb: bool, init_state, engine):
    if param_ids is None:
        param_ids = range(len(ex_ops))
    logger.info(f"Entering `get_energy`")
    ci_strings = get_ci_strings(n_qubits, n_elec, hcb)
    init_state = translate_init_state(init_state, n_qubits, ci_strings)
    ket = GETVECTOR_MAP[engine](
        params, n_qubits, n_elec, tuple(ex_ops), tuple(param_ids), hcb=hcb, init_state=init_state
    )
    hket = apply_op(hamiltonian, ket)
    return ket @ hket


get_energy_statevector = partial(get_energy, engine="statevector")
try:
    get_energy_and_grad_statevector = jit(value_and_grad(get_energy_statevector), static_argnums=[2, 3, 4, 5, 6])
except NotImplementedError:

    def get_energy_and_grad_statevector(*args, **kwargs):
        raise NotImplementedError("Non JAX-backend for statevector engine")


get_energy_tensornetwork = partial(get_energy, engine="tensornetwork")
try:
    get_energy_and_grad_tensornetwork = jit(value_and_grad(get_energy_tensornetwork), static_argnums=[2, 3, 4, 5, 6])
except NotImplementedError:

    def get_energy_and_grad_tensornetwork(*args, **kwargs):
        raise NotImplementedError("Non JAX-backend for tensornetwork engine")


ENERGY_AND_GRAD_MAP = {
    "tensornetwork": get_energy_and_grad_tensornetwork,
    "statevector": get_energy_and_grad_statevector,
    "civector": get_energy_and_grad_civector,
    "civector-large": get_energy_and_grad_civector_nocache,
    "pyscf": get_energy_and_grad_pyscf,
}


def get_energy_and_grad(params, hamiltonian, n_qubits, n_elec, ex_ops, param_ids, hcb, init_state, engine):
    if engine not in ENERGY_AND_GRAD_MAP:
        raise ValueError(f"Engine '{engine}' not supported")

    func = ENERGY_AND_GRAD_MAP[engine]
    ci_strings = get_ci_strings(n_qubits, n_elec, hcb)
    init_state = translate_init_state(init_state, n_qubits, ci_strings)
    return func(params, hamiltonian, n_qubits, n_elec, tuple(ex_ops), tuple(param_ids), hcb, init_state)


APPLY_EXCITATION_MAP = {
    # share the same function with statevector engine
    "tensornetwork": apply_excitation_statevector,
    "statevector": apply_excitation_statevector,
    "civector": apply_excitation_civector,
    "civector-large": apply_excitation_civector_nocache,
    "pyscf": apply_excitation_pyscf,
}


def apply_excitation(state, n_qubits, n_elec, ex_op, hcb, engine):
    if engine not in APPLY_EXCITATION_MAP:
        raise ValueError(f"Engine '{engine}' not supported")

    state = tc.backend.convert_to_tensor(state)

    is_statevector_input = len(state) == (1 << n_qubits)
    is_statevector_engine = engine in ["tensornetwork", "statevector"]

    if is_statevector_input and not is_statevector_engine:
        ci_strings = get_ci_strings(n_qubits, n_elec, hcb)
        state = statevector_to_civector(state, ci_strings)
    if not is_statevector_input and is_statevector_engine:
        ci_strings = get_ci_strings(n_qubits, n_elec, hcb)
        state = civector_to_statevector(state, n_qubits, ci_strings)
    func = APPLY_EXCITATION_MAP[engine]
    res_state = func(state, n_qubits, n_elec, ex_op, hcb)
    if is_statevector_input and not is_statevector_engine:
        return civector_to_statevector(res_state, n_qubits, ci_strings)
    if not is_statevector_input and is_statevector_engine:
        return statevector_to_civector(res_state, ci_strings)
    return res_state


def translate_init_state(init_state, n_qubits, ci_strings):
    if init_state is None:
        return None
    # translate to civector first for all engines to be JAX-compatible
    if isinstance(init_state, tc.Circuit):
        # note no cupy backend for tc
        init_state = statevector_to_civector(tc.backend.convert_to_tensor(init_state.state().real), ci_strings)
    else:
        is_statevector_input = len(init_state) == (1 << n_qubits)
        if is_statevector_input:
            init_state = statevector_to_civector(init_state, ci_strings)
    return tc.backend.convert_to_tensor(init_state)



#  Copyright (c) 2023. The TenCirChem Developers. All Rights Reserved.
#
#  This file is distributed under ACADEMIC PUBLIC LICENSE
#  and WITHOUT ANY WARRANTY. See the LICENSE file for details.


from typing import Tuple, Any
import logging

import numpy as np
from pyscf.fci import cistring
from pyscf.fci.addons import des_a, cre_a, des_b, cre_b
import tensorcircuit as tc

from tencirchem.static.hamiltonian import apply_op
from tencirchem.static.ci_utils import get_init_civector

Tensor = Any

logger = logging.getLogger(__name__)


class CIvectorPySCF:
    def __init__(self, civector, n_orb, n_elec_a, n_elec_b):
        assert isinstance(civector, np.ndarray)
        self.civector = civector.reshape(-1)
        self.n_orb = n_orb
        self.n_elec_a = n_elec_a
        self.n_elec_b = n_elec_b

    def cre(self, i):
        n_elec_a, n_elec_b = self.n_elec_a, self.n_elec_b
        if i < self.n_orb:
            cre = cre_a
            n_elec_a += 1
        else:
            cre = cre_b
            n_elec_b += 1

        new_civector = cre(self.civector, self.n_orb, (self.n_elec_a, self.n_elec_b), i % self.n_orb)
        return CIvectorPySCF(new_civector, self.n_orb, n_elec_a, n_elec_b)

    def des(self, i):
        n_elec_a, n_elec_b = self.n_elec_a, self.n_elec_b
        if i < self.n_orb:
            des = des_a
            n_elec_a -= 1
        else:
            des = des_b
            n_elec_b -= 1

        new_civector = des(self.civector, self.n_orb, (self.n_elec_a, self.n_elec_b), i % self.n_orb)
        return CIvectorPySCF(new_civector, self.n_orb, n_elec_a, n_elec_b)

    def pq(self, p, q):
        return self.des(q).cre(p)

    def pqqp(self, p, q):
        return self.des(p).cre(q).des(q).cre(p)

    def pqrs(self, p, q, r, s):
        return self.des(s).des(r).cre(q).cre(p)

    def pqrssrqp(self, p, q, r, s):
        return self.des(p).des(q).cre(r).cre(s).des(s).des(r).cre(q).cre(p)


def apply_a2_pyscf(civector: CIvectorPySCF, ex_op) -> Tensor:
    if len(ex_op) == 2:
        apply_f = civector.pqqp
    else:
        assert len(ex_op) == 4
        apply_f = civector.pqrssrqp
    civector1 = apply_f(*ex_op)
    civector2 = apply_f(*reversed(ex_op))
    return -civector1.civector - civector2.civector


def apply_a_pyscf(civector: CIvectorPySCF, ex_op) -> Tensor:
    if len(ex_op) == 2:
        apply_func = civector.pq
    else:
        assert len(ex_op) == 4
        apply_func = civector.pqrs
    civector1 = apply_func(*ex_op)
    civector2 = apply_func(*reversed(ex_op))
    return civector1.civector - civector2.civector


def evolve_excitation_pyscf(civector: Tensor, ex_op, n_orb, n_elec, theta) -> Tensor:
    ket = CIvectorPySCF(civector, n_orb, n_elec // 2, n_elec // 2)
    aket = apply_a_pyscf(ket, ex_op)
    a2ket = apply_a2_pyscf(ket, ex_op)
    return civector + (1 - np.cos(theta)) * a2ket + np.sin(theta) * aket


def get_civector_pyscf(params, n_qubits, n_elec, ex_ops, param_ids, hcb=False, init_state=None):
    assert not hcb
    n_orb = n_qubits // 2
    num_strings = cistring.num_strings(n_orb, n_elec // 2) ** 2

    if init_state is None:
        civector = get_init_civector(num_strings)
    else:
        civector = tc.backend.convert_to_tensor(init_state)

    civector = tc.backend.numpy(civector)

    for ex_op, param_id in zip(ex_ops, param_ids):
        theta = params[param_id]
        civector = evolve_excitation_pyscf(civector, ex_op, n_orb, n_elec, theta)

    return civector.reshape(-1)


def get_energy_and_grad_pyscf(
    params, hamiltonian, n_qubits, n_elec, ex_ops: Tuple, param_ids: Tuple, hcb: bool = False, init_state=None
):
    params = tc.backend.numpy(params)
    ket = get_civector_pyscf(params, n_qubits, n_elec, ex_ops, param_ids, hcb, init_state)
    bra = tc.backend.numpy(apply_op(hamiltonian, ket))
    energy = bra @ ket

    gradients_beforesum = _get_gradients_pyscf(bra, ket, params, n_qubits, n_elec, ex_ops, param_ids, hcb)

    gradients = np.zeros(params.shape)
    for grad, param_id in zip(gradients_beforesum, param_ids):
        gradients[param_id] += grad

    return energy, 2 * gradients


def _get_gradients_pyscf(bra, ket, params, n_qubits, n_elec, ex_ops, param_ids, hcb):
    assert not hcb

    n_orb = n_qubits // 2

    gradients_beforesum = []
    for param_id, ex_op in reversed(list(zip(param_ids, ex_ops))):
        theta = params[param_id]
        bra = evolve_excitation_pyscf(bra, ex_op, n_orb, n_elec, -theta)
        ket = evolve_excitation_pyscf(ket, ex_op, n_orb, n_elec, -theta)
        ket_pyscf = CIvectorPySCF(ket, n_orb, n_elec // 2, n_elec // 2)
        fket = apply_a_pyscf(ket_pyscf, ex_op)
        grad = bra @ fket
        gradients_beforesum.append(grad)
    gradients_beforesum = list(reversed(gradients_beforesum))
    gradients_beforesum = np.array(gradients_beforesum)

    return gradients_beforesum


def apply_excitation_pyscf(civector, n_qubits, n_elec, f_idx, hcb):
    assert not hcb
    civector_pyscf = CIvectorPySCF(civector, n_qubits // 2, n_elec // 2, n_elec // 2)
    return apply_a_pyscf(civector_pyscf, f_idx)
